from RoboidControl.Participant import Participant
from RoboidControl.Thing import Thing
from RoboidControl.Messages import *
from RoboidControl.Messages.BinaryMsg import BinaryMsg
from RoboidControl.Messages.ParticipantMsg import ParticipantMsg
from RoboidControl.Messages.NetworkIdMsg import NetworkIdMsg
from RoboidControl.Messages.ThingMsg import ThingMsg
from RoboidControl.Messages.NameMsg import NameMsg
from RoboidControl.Messages.ModelUrlMsg import ModelUrlMsg
from RoboidControl.Messages.IMessage import IMessage
from RoboidControl.Messages.InvestigateMsg import InvestigateMsg
from RoboidControl.Messages.PoseMsg import PoseMsg

import socket
import threading
import time
from typing import Optional
import logging
logger = logging.getLogger(__name__)

# import sys
# micropython = 'micropython' in sys.modules

# if micropython:
#     from MicroPython.uPythonParticipant import uPythonParticipant

class ParticipantUDP(Participant):
    """! A participant using UDP communication
    
    A local participant is the local device which can communicate with
    other participants It manages all local things and communcation with other
    participants. Each application has a local participant which is usually
    explicit in the code. An participant can be isolated. In that case it is
    standalong and does not communicate with other participants.
    
    It is possible to work with an hidden participant by creating things without
    specifying a participant in the constructor. In that case an hidden isolated
    participant is created which can be obtained using
    RoboidControl::IsolatedParticipant::Isolated().
    @sa RoboidControl::Thing::Thing()
    """

    name = "Participant"

#region Init
    def __init__(self, 
                 port: int = 7681,
                 ip_address: Optional[str] = None,
                 local_port: int = 7681
                 ) -> None:
        """! Create a participant
        @param ipAddress The IP address of the site
        @param port The port used by the site
        @param localPort The port used by the local participant
        """
        super().__init__(ip_address = "127.0.0.1", port = local_port)

        ## True if the participant is running isolated.
        # Isolated participants do not communicate with other participants
        self.is_isolated: bool = True
        ## The remote site when this participant is connected to a site
        self.remote_site: Optional[Participant] = None

        ## The interval in milliseconds for publishing (broadcasting) data on the local network
        self.publishInterval = 3000 # 3 seconds

        if port != 0:
            self.is_isolated = False
            if ip_address is not None:
                self.remote_site = Participant(ip_address, port)

        Participant.AddParticipant(self)

        self.buffer: bytearray = bytearray(256)

        self.udp_socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)        
        self.udp_socket.setsockopt(socket.SOL_SOCKET, socket.SO_BROADCAST, 1) 
        self.udp_socket.bind(("0.0.0.0", local_port))

        self.thread = threading.Thread(target = self.Receiver)
        self.thread.daemon = True
        self.thread.start()
    
    @staticmethod
    def Isolated() -> 'ParticipantUDP':
        if ParticipantUDP.isolated_participant == None:
            ParticipantUDP.isolated_participant = ParticipantUDP(0)
        return ParticipantUDP.isolated_participant

#endregion Init

    isolated_participant = None

#region Update

    def Update(self):
        currentTimeMs = int(time.time() * 1000)

        if self.is_isolated == False:
            if self.publishInterval > 0 and currentTimeMs > self.nextPublishMe:
                msg = ParticipantMsg(self.network_id)
                if self.remote_site == None:
                    self.Publish(msg)
                else:
                    self.Send(self.remote_site, msg)
                    
                self.nextPublishMe = currentTimeMs + self.publishInterval

        self.UpdateMyThings(currentTimeMs)

    nextPublishMe = 0

    def UpdateMyThings(self, currentTimeMs: int):
        for thing in self.things:
            # if thing.hierarchyChanged and not (self.isIsolated or self.network_id == ):
            #     thingMsg = ThingMsg(self.network_id, thing)
            #     self.Send(self.remote_site, thingMsg)
            if thing.owner is not None:
                poseMsg = PoseMsg(thing.owner.network_id, thing)
                if self.remote_site is not None:
                    self.Send(self.remote_site, poseMsg)

            thing.Update(False)
            if not(self.is_isolated or self.network_id == 0):
                if thing.terminate:
                    destroyMsg = DestroyMsg.Create(self.network_id, thing)
                    self.Send(self.remote_site, destroyMsg)
                else:
                    if self.remote_site is not None and thing.owner is not None:
                        # Send to remote site
                        self.Send(self.remote_site, PoseMsg(thing.owner.network_id, thing))
                        self.Send(self.remote_site, BinaryMsg(thing.owner.network_id, thing))
            if thing.terminate:
                self.Remove(thing)

#endregion

#region Send

    def SendThingInfo(self, owner: Participant, thing: 'Thing', recursively: bool = False):
        self.Send(owner, ThingMsg(self.network_id, thing))
        self.Send(owner, NameMsg(self.network_id, thing))
        self.Send(owner, ModelUrlMsg(self.network_id, thing))
        self.Send(owner, PoseMsg(self.network_id, thing, True))
        self.Send(owner, BinaryMsg(self.network_id, thing))

        if recursively:
            for child in thing.children:
                self.SendThingInfo(owner, child, recursively)

    def PublishThingInfo(self, thing: Thing):
        self.Publish(ThingMsg(self.network_id, thing))
        self.Publish(NameMsg(self.network_id, thing))
        self.Publish(ModelUrlMsg(self.network_id, thing))
        self.Publish(PoseMsg(self.network_id, thing))
        self.Publish(BinaryMsg(self.network_id, thing))

    def Send(self, owner: Participant, msg: IMessage):
        buffer_size = msg.Serialize([self.buffer])
        if buffer_size <= 0:
            return True
        
        # print(f'{self.name} send {self.buffer[0]} to {owner.ip_address} {owner.port}')
        self.udp_socket.sendto(self.buffer[:buffer_size], (owner.ip_address, owner.port))
        return True

    def Publish(self, msg: IMessage):
        buffer_size = msg.Serialize([self.buffer])
        if buffer_size <= 0:
            return True
        
        # print(f'publish {self.buffer[0]} to {self.port}')
        self.udp_socket.sendto(self.buffer[:buffer_size], ('<broadcast>', self.port))
        return True

#endregion       

#region Receive
    
    def Receiver(self):
        while True:
            try:
                data, addr = self.udp_socket.recvfrom(1024)
                remote_ip_address = addr[0]
                remote_port = addr[1]
                # print(f'msg received from {remote_ip_address}:{remote_port}')
                remote_participant = Participant.GetParticipant(remote_ip_address, remote_port)
                if remote_participant is None:
                    # print(f'new participant')
                    remote_participant = Participant.AddParticipant(remote_ip_address, remote_port)
                self.ReceiveData(data, remote_participant)
            except ConnectionError:
                self.network_id = 0
                pass

    def ReceiveData(self, data: bytes, sender: Participant):
        msgId = data[0]
        # print(f'msg {msgId} ')
        match msgId:
            case ParticipantMsg.id:
                self.ProcessParticipantMsg(sender, ParticipantMsg(data))
            case NetworkIdMsg.id:
                self.ProcessSiteIdMsg(sender, NetworkIdMsg(data))
            # case InvestigateMsg.id:
            #     self.ProcessInvestigateMsg(InvestigateMsg(data))
            case ThingMsg.id:
                self.ProcessThingMsg(sender, ThingMsg(data))
            case NameMsg.id:
                self.ProcessNameMsg(sender, NameMsg(data))
            case ModelUrlMsg.id:
                self.ProcessModelUrlMsg(sender, ModelUrlMsg(data))
            case BinaryMsg.id:
                self.ProcessBinaryMsg(sender, BinaryMsg(data))            
            case _:
                pass

    def ProcessParticipantMsg(self, sender: Participant, msg: ParticipantMsg):
        logger.debug(f'{self.name} Process participantMsg {msg.network_id}')

    def ProcessSiteIdMsg(self, sender: Participant, msg: NetworkIdMsg):
        logger.debug(f'{self.name} Process NetworkIdMsg {self.network_id} -> {msg.network_id}')
        if self.network_id != msg.network_id:
            self.network_id = msg.network_id
            for thing in self.things:
                #if thing.parent is None:
                self.SendThingInfo(sender, thing, recursively=True)

    def ProcessInvestigateMsg(self, sender: Participant, msg: InvestigateMsg):
        logger.debug(f'{self.name} Process InestigateMsg [{msg.network_id}/{msg.thing_id}]')

    def ProcessThingMsg(self, sender: Participant, msg: ThingMsg):
        logger.debug(f'{self.name}: Process ThingMsg [{msg.network_id}/{msg.thing_id}] {msg.thing_type} {msg.parent_id}')

    def ProcessNameMsg(self, sender: Participant, msg: NameMsg):
        logger.debug(f'{self.name}: Process NameMsg [{msg.network_id}/{msg.thing_id}] {msg.name_length} {msg.name}')

        owner: Optional[Participant] = Participant.GetParticipant(msg.network_id)
        if owner is None:
            return
        
        thing: Optional[Thing] = self.Get(msg.thing_id)
        if thing is None:
            return

        thing.name = msg.name
    
    def ProcessModelUrlMsg(self, sender: Participant, msg: ModelUrlMsg):
        logger.debug(f'{self.name}: Process ModelUrlMsg [{msg.network_id}/{msg.thing_id}] {msg.url_length} {msg.url}')

    def ProcessPoseMsg(self, sender: Participant, msg: PoseMsg):
        logger.debug(f'{self.name}: Process PoseMsg [{msg.network_id}/{msg.thing_id}] {msg.pose_type}')

        owner: Optional[Participant] = Participant.GetParticipant(msg.network_id)
        if owner is None:
            return
        
        thing: Optional[Thing] = self.Get(msg.thing_id)
        if thing is None:
            return

        if (msg.pose_type & PoseMsg.Position) != 0:
            thing.position = msg.position
        if (msg.pose_type & PoseMsg.Orientation) != 0:
            thing.orientation = msg.orientation
        if (msg.pose_type & PoseMsg.LinearVelocity) != 0:
            thing.linear_velocity = msg.linear_velocity
        if (msg.pose_type & PoseMsg.AngularVelocity) != 0:
            thing.angular_velocity = msg.angular_velocity

    def ProcessBinaryMsg(self, sender: Participant, msg: BinaryMsg):
        logger.debug(f'{self.name}: Process BinaryMsg [{msg.network_id}/{msg.thing_id}] {msg.data_length}')

        thing: Optional[Thing] = self.Get(msg.thing_id)
        if thing is not None:
            thing.ProcessBinary(msg.data)

#endregion